import{p as e,f as o}from"./flowDb-956e92f1-94f38b83.js";import{f as t,g as a}from"./styles-c10674c1-9588494e.js";import{u as i}from"./index-cce6f8a1.js";import"./graph-0d0a2c10.js";import"./layout-da885b9b.js";import"./index-3862675e-58ea0305.js";import"./clone-fdf164e2.js";import"./edges-e0da2a9e-f837ca8a.js";import"./createText-2e5e7dd3-ca47fd38.js";import"./line-f1c817d3.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-f1efda17.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
